#pragma rtGlobals=1		// Use modern global access method.

//	Touchstone file loader
//	The Touchstone file format was originated by Agilent Corporation to store n-port parameter data.
//	This file loader is based on the following draft specification:
//		http://www.eda.org/pub/ibis/connector/touchstone_spec11.pdf

Constant kTouchStoneFileLoaderVersion = 1.00

// Add an item to Igor's Load Waves submenu (Data menu)
Menu "Load Waves"
	"Load Touchstone File...", /Q, LoadTouchstoneFileMenuItem()
	"Load All Touchstone Files In Directory...", /Q, LoadAllTouchstoneFilesMenuItem()
End

Menu "Macros"
	"View Touchstone Data From One File", /Q, ViewTouchstoneFileMenuItem()
	"View Touchstone Data From Multiple Files", /Q, ViewTouchstoneFilesMenuItem()
	"Kill All Touchstone Windows", /Q, KillAllTouchstoneWindows()
	"View Touchstone Help", /Q, DisplayHelpTopic "Touchstone File Loader"
End

// These are used when displaying data from one file only.
static Constant kGraphModeUndefined = -1
static Constant kGraphModeAllInOne = 1					// Display all values in a single graph
static Constant kGraphModeValuesSeparate = 2			// Display each value in a separate graph
static Constant kGraphModePairsSeparate = 3			// Display each value pair in a separate graph

Structure TouchstoneFileInfo
	Variable numberOfPorts				// 1, 2, 3, 4 . . . This is deduced from the file name.

	// These come from the option line which (for example) looks something like this:
	//		# GHZ S DB R 50
	String frequencyUnit				// "GHZ", "MHZ", "KHZ", etc.
	String parameterKind				// "S" (S-parameter), "Y" (admittance),"Z" (Z-parameter), "H" (hybrid), "G" (hybrid-G)
	String format						// "DB" (dB-angle), "MA" (magnitude-angle), "RI" (real-imaginary)
	Variable reference					// Reference resistance in ohms
EndStructure

static Function StoreTouchstoneFileInfo(fileInfo)		// Stores info as variables in current data folder.
	STRUCT TouchstoneFileInfo &fileInfo

	Variable/G numberOfPorts = fileInfo.numberOfPorts
	String/G frequencyUnit = fileInfo.frequencyUnit
	String/G parameterKind = fileInfo.parameterKind
	String/G format = fileInfo.format
	Variable/G reference = fileInfo.reference
End

static Function RetrieveTouchstoneFileInfo(fileInfo)		// Retrives info from variables in current data folder.
	STRUCT TouchstoneFileInfo &fileInfo
	
	NVAR numberOfPorts, reference
	SVAR frequencyUnit, parameterKind, format

	fileInfo.numberOfPorts = numberOfPorts
	fileInfo.frequencyUnit = frequencyUnit
	fileInfo.parameterKind = parameterKind
	fileInfo.format = format
	fileInfo.reference = reference
End

static Function/S GetNextSymbol(text, pos)
	String text
	Variable &pos						// Input and output
	
	String symbol = ""
	
	Variable textLen = strlen(text)
	
	Variable ch

	do
		ch = char2num(text[pos])
		if (ch > 0x20)					// Past all white space and control characters?
			break
		endif
		pos += 1
	while(pos < textLen)

	// Now deposit characters into symbol.
	if (pos < textLen)
		do
			ch = char2num(text[pos])
			
			if (ch <= 0x20)												// End of symbol?
				break
			endif
			
			symbol += text[pos]
			pos += 1
		while(pos < textLen)
	endif

	return symbol
End

static Function/S ParseOptionLine(optionStr, fileInfo, quiet)
	String optionStr
	STRUCT TouchstoneFileInfo &fileInfo
	Variable quiet

	fileInfo.frequencyUnit = "GHZ"
	fileInfo.parameterKind = "S"
	fileInfo.format = "MA"
	fileInfo.reference = 50

	// Symbols in the option line can appear in any order which greatly complicates interpreting it
	Variable pos = 1					// 1 to skip the initial # symbol
	do
		String symbol = GetNextSymbol(optionStr, pos)
		if (strlen(symbol) == 0)
			break
		endif
		
		strswitch(symbol)
			case "GHZ":
				fileInfo.frequencyUnit = symbol
					break
			
			case "MHZ":
				fileInfo.frequencyUnit = symbol
				break

			case "KHZ":
				fileInfo.frequencyUnit = symbol
				break

			case "HZ":
				fileInfo.frequencyUnit = symbol
				break
			
			case "S":
			case "Y":
			case "Z":
			case "H":
			case "G":
				fileInfo.parameterKind = symbol
				break
				
			case "DB":
			case "MA":
			case "RI":
				fileInfo.format = symbol
				break
				
			case "R":
				symbol = GetNextSymbol(optionStr, pos)
				fileInfo.reference = str2num(symbol)
				break

			case "!":			// added by Alice Tsui to also work with SDD ONLY PLTS exported files
				break			

			default:
				if (quiet != 2)
					Printf "Unknown symbol in option line: \"%s\"\r", symbol
				endif
				break
		endswitch	
	while(1)
End

static Function ExtractTouchstoneOptionInfo(refNum, fileInfo, quiet)
	Variable refNum			// File reference number
	STRUCT TouchstoneFileInfo &fileInfo
	Variable quiet				// 0 = print diagnostic and error messages, 1 = print error messages only, 2 = do not print diagnostic or error messages
	
	Variable result = 0
	
	// Set defaults used if option line is missing.
	fileInfo.frequencyUnit = "GHZ"
	fileInfo.parameterKind= "S"
	fileInfo.format = "MA"
	fileInfo.reference = 50
	
	// Read option string
	String optionStr = ""
	FSetPos refNum, 0
	Variable done = 0
	do
		FReadLine refNum, optionStr
		if (strlen(optionStr) == 0)
			Close refNum
			return -1								// Unexpected end of file
		endif
		
		strswitch(optionStr[0])
			case "#":								// Option line?
				ParseOptionLine(optionStr, fileInfo, quiet)
				done = 1
				break
			
			case "!":								// Comment line?
				break
			
			default:									// Should be data line.
				done = 1
				break
		endswitch
	while(!done)
	
	return result
End

static Function SkipOptionAndCommentLines(refNum, dataStartLine, dataStartPos, quiet)
	Variable refNum			// File reference number
	Variable &dataStartLine
	Variable &dataStartPos		// Start of data as file position relative to start of file
	Variable quiet				// 0 = print diagnostic and error messages, 1 = print error messages only, 2 = do not print diagnostic or error messages
	
	dataStartLine = 0
	dataStartPos = 0
	
	Variable result = 0
	
	String line
	Variable done = 0
	do
		FStatus refNum
		dataStartPos = V_filePos
		FReadLine refNum, line
		if (strlen(line) == 0)
			Close refNum
			return -1								// Unexpected end of file
		endif
		
		strswitch(line[0])
			case "#":								// Option line?
				dataStartLine +=  1
				break
			
			case "!":								// Comment line?
				dataStartLine +=  1
				break
			
			default:									// Should be data line.
				FSetPos refNum, dataStartPos
				done = 1
				break
		endswitch
	while(!done)
	
	return result
End

static Function SetTouchstoneViewPrefs(createTable, createGraph, graphMode)
	Variable createTable, createGraph, graphMode
	
	NewDataFolder/O root:Packages
	NewDataFolder/O root:Packages:TouchstoneLoader
	
	Variable/G root:Packages:TouchstoneLoader:createTable = createTable
	Variable/G root:Packages:TouchstoneLoader:createGraph = createGraph
	Variable/G root:Packages:TouchstoneLoader:graphMode = graphMode
End

static Function GetTouchstoneViewPrefs(createTable, createGraph, graphMode)
	Variable &createTable, &createGraph, &graphMode						// Outputs
	
	NewDataFolder/O root:Packages
	NewDataFolder/O root:Packages:TouchstoneLoader
	
	createTable = 1
	NVAR/Z gCreateTable = root:Packages:TouchstoneLoader:createTable
	if (NVAR_Exists(gCreateTable))
		createTable =  gCreateTable
	endif
	if ( (createTable<0) || (createTable>3) )
		createTable = 1
	endif
	
	createGraph = 1
	NVAR/Z gCreateGraph = root:Packages:TouchstoneLoader:createGraph
	if (NVAR_Exists(gCreateGraph))
		createGraph =  gCreateGraph
	endif
	if ( (createGraph<0) || (createGraph>3) )
		createGraph = 1
	endif
	
	graphMode = kGraphModeAllInOne
	NVAR/Z gGraphMode = root:Packages:TouchstoneLoader:graphMode
	if (NVAR_Exists(gGraphMode))
		graphMode =  gGraphMode
	endif
	if ( (graphMode<kGraphModeAllInOne) || (graphMode>kGraphModePairsSeparate) )
		graphMode = kGraphModeAllInOne
	endif
End

//	ViewTouchstoneFileDialog(title, createTable, createGraph, graphMode)
//	Displays a dialog in which the user can enter parameters.
//	This is called when we are displaying Touchstone data from a single file.
//	createTable, createGraph and graphMode are outputs only.
//	The function result is 0 if OK or -1 if cancel.
Function ViewTouchstoneFileDialog(title, createTable, createGraph, graphMode)
	String title					// Dialog title
	Variable &createTable		// Output. 0 = no table, 1 = edit even value column, 2 = edit odd value columns, 3 = edit both
	Variable &createGraph		// Output. 0 = no graph, 1 = graph even value column, 2 = graph odd value columns, 3 = graph both
	Variable &graphMode		// Output. kGraphModeAllInOne, kGraphModeValuesSeparate or kGraphModePairsSeparate

	GetTouchstoneViewPrefs(createTable, createGraph, graphMode)
	
	Variable createTable2 = createTable + 1
	Variable createGraph2 = createGraph + 1
	Variable graphMode2 = graphMode
	
	Prompt createTable2, "Create table", popup "No table;Even value columns;Odd value columns;All value columns;"
	Prompt createGraph2, "Create graph", popup "No graph;Even value columns;Odd value columns;All value columns;"
	Prompt graphMode2, "Graph mode", popup "All values in one graph;Each value in a separate graph;Each pair of values in a separate graph;"
	DoPrompt title, createTable2, createGraph2, graphMode2
	if (V_Flag != 0)
		return -1			// User cancelled.
	endif
	
	createTable = createTable2 - 1
	createGraph = createGraph2 - 1
	graphMode = graphMode2

	SetTouchstoneViewPrefs(createTable, createGraph, graphMode)
	
	return 0
End

static Function/S TouchstoneFreqWaveName(frequencyUnit)
	String frequencyUnit

//	String freqName= "freq_" + frequencyUnit		// e.g., freq_GHZ
	String freqName= "X_freq_" + frequencyUnit	// modified by Alice Tsui
	return freqName
End

static Function/S TouchstoneValueWaveName(numberOfPorts, parameterKind, format, valueIndex)
	Variable numberOfPorts
	String parameterKind
	String format
	Variable valueIndex				// Zero-based index into the columns of parameter values.
	
	String suffix, name
	
	if (valueIndex >= 2*numberOfPorts^2)
		return "Bad valueIndex parameter"					// Should never happen
	endif

	Variable index
	Variable first
	Variable second

	index = trunc(valueIndex / 2)
	if (numberOfPorts <= 2)
		first = mod(index,numberOfPorts) + 1
		second = trunc(index/numberOfPorts)+1
	else
		first = trunc(index/numberOfPorts)+1
		second = mod(index,numberOfPorts) + 1
	endif

	suffix = num2istr(first) + num2istr(second)
	
//	name = parameterKind + suffix
	name = "X_" + parameterKind + suffix	// modified by Alice Tsui
//	name = "X_Sdd" + suffix	// modified by Alice Tsui
		
	strswitch(format)
		case "DB":
			if (valueIndex & 1)
				name += "_A"
			else
				name += "_DB"
			endif
			break
			
		case "MA":
			if (valueIndex & 1)
				name += "_A"
			else
				name += "_M"
			endif
			break
			
		case "RI":
			if (valueIndex & 1)
				name += "_I"
			else
				name += "_R"
			endif
			break
			
		case "":
			break					// Caller does not want format appended.
	endswitch
	
	return name
End

static Function SkipWhiteSpace(buffer, bufferLen, pos)
	String buffer
	Variable bufferLen
	Variable &pos						// Input and output
	
	Variable ch

	if (pos < bufferLen)
		do
			ch = char2num(buffer[pos])
			
			if (ch == 0x21)											// Start of a comment?
				pos = bufferLen											// Skip the rest of the line.
				break
			endif
			
			if (ch==0x2B || ch==0x2D || ch >= 0x30)					// A plus, minus or digit will pass this test
				break
			endif
			pos += 1					// Skip space, tab, CR or LF
		while(pos < bufferLen)
	endif
End

static Function ReadValue(buffer, bufferLen, pos, val)
	String buffer
	Variable bufferLen
	Variable &pos						// Input and output
	Variable &val						// Output
	
	Variable err = 0
	
	val = str2num(buffer[pos, bufferLen])
	
	Variable ch

	// Now skip this number
	do
		ch = char2num(buffer[pos])
		if (ch==0x20 || ch==0x09 || ch==0x0D || ch==0x0A)			// A space, tab, CR or LF will pass this test
			break
		endif
		pos += 1
	while(pos < bufferLen)
	
	SkipWhiteSpace(buffer, bufferLen, pos)
	
	return err
End

static Function FindXScaling(freq, x0, dx)
	Wave freq
	Variable &x0, &dx				// Outputs
	
	x0 = freq[0]
	dx = freq[1] - freq[0]
	
	Variable epsilon = abs(x0 / 10000)
	
	// Now make sure that frequencies are really evenly spaced.
	Variable numPoints = numpnts(freq)
	Variable i
	for(i=1; i<numPoints-2; i+=1)
		Variable currentDX = freq[i+1]-freq[i]
		if (abs(currentDX-dx) > epsilon)
			x0 = 0						// Signal that frequencies are not evenly-spaced
			dx = 0
		endif	
	endfor
End

static Function LoadTouchstoneData(fileInfo, refNum, dataStartPos, dataStartLine, quiet)
	STRUCT TouchstoneFileInfo &fileInfo
	Variable refNum
	Variable dataStartPos			// Start of data as file position relative to start of file
	Variable dataStartLine			// Zero-based
	Variable quiet					// 0 = print diagnostic and error messages, 1 = print error messages only, 2 = do not print diagnostic or error messages

	String name
	
	Variable numberOfFrequencies = 0
	Variable numberOfPointsPerWave =  1024				// Initial size of waves. They grow as needed.

	Variable numberOfValuesPerFreq
	numberOfValuesPerFreq = (fileInfo.numberOfPorts ^ 2) * 2
	
	// Make the frequency wave
	String freqName= TouchstoneFreqWaveName(fileInfo.frequencyUnit)		// e.g., freq_GHZ
	Make/O/N=(numberOfPointsPerWave) $freqName
	
	// Make the value waves
	Variable valueColumnIndex
	for(valueColumnIndex=0; valueColumnIndex<numberOfValuesPerFreq; valueColumnIndex+=1)
		name = TouchstoneValueWaveName(fileInfo.numberOfPorts, fileInfo.parameterKind, fileInfo.format, valueColumnIndex)
		Make/O/N=(numberOfPointsPerWave) $name
	endfor

	Variable lineNumber = dataStartLine - 1
	Variable pointNum						// Point we are currently working on.
	Variable waveIndex						// -1 for frequency wave, 0 for s11, 1 for s12, etc.
	String buffer
	Variable bufferLen
	Variable pos							// Next position in buffer to read from.
	Variable val
	Variable err = 0
	
	buffer = ""
	bufferLen = 0
	pos = 0
	waveIndex = -1
	pointNum = 0
	do
		if (pos >= bufferLen)
			do												// Read till we get a non-blank line.
				lineNumber += 1
				
				FReadLine refNum, buffer
				bufferLen = strlen(buffer)
				if (bufferLen == 0)
					break									// Hit end of file.
				endif
			
				if (CmpStr(buffer[0], "!") == 0)			// Is this a comment line?
					if (strsearch(buffer, "Noise Parameters", 0) > 0)	// Start of optional noise parameters?
						bufferLen = 0									// Signal end of file.
						break											// We do not support loading of noise parameters at this time.
					endif
					continue								// Skip comment line.
				endif
			
				pos = 0
				SkipWhiteSpace(buffer, bufferLen, pos)		// Skip any leading white space.
				if (pos < bufferLen)							// Got some non-white space characters?
					break
				endif
			while(1)
		endif

		if (bufferLen == 0)
			break											// Hit end of file.
		endif
		
		err = ReadValue(buffer, bufferLen, pos, val)
		if (err != 0)
			break
		endif
		
		// Check if we need to increase size of waves
		if (pointNum >= numberOfPointsPerWave)
			numberOfPointsPerWave += 1024
			Wave w = $freqName
			Redimension/N=(numberOfPointsPerWave) w
			for(valueColumnIndex=0; valueColumnIndex<numberOfValuesPerFreq; valueColumnIndex+=1)
				name = TouchstoneValueWaveName(fileInfo.numberOfPorts, fileInfo.parameterKind, fileInfo.format, valueColumnIndex)
				Wave w = $name
				Redimension/N=(numberOfPointsPerWave) w
			endfor
		endif
		
		if (waveIndex < 0)
			Wave w = $freqName
			w[pointNum] = val
			waveIndex = 0
		else
			name = TouchstoneValueWaveName(fileInfo.numberOfPorts, fileInfo.parameterKind, fileInfo.format, waveIndex)
			Wave w = $name
			w[pointNum] = val
			waveIndex += 1
			if (waveIndex >= numberOfValuesPerFreq)
				waveIndex = -1
				pointNum += 1
			endif
		endif
	while(1)
	
	Variable numPointsLoaded = pointNum + 1
	if (waveIndex < 0)
		// pointNum is number of  NEXT point to be loaded
		numPointsLoaded -= 1
	endif

	// Delete any unsed points in waves
	if (numberOfPointsPerWave > numPointsLoaded)
		Wave w = $freqName
		Redimension/N=(numPointsLoaded) w
		for(valueColumnIndex=0; valueColumnIndex<numberOfValuesPerFreq; valueColumnIndex+=1)
			name = TouchstoneValueWaveName(fileInfo.numberOfPorts, fileInfo.parameterKind, fileInfo.format, valueColumnIndex)
			Wave w = $name
			Redimension/N=(numPointsLoaded) w
		endfor
	endif
	
	// Set wave scaling and units

	SetScale /P d 0, 0, ""+fileInfo.frequencyUnit, $freqName	// Set data units of frequency wave
	
	Variable x0, dx												// Set X units of value waves if frequencies are evenly spaced.
	FindXScaling($freqName, x0, dx)
	if (dx != 0)													// Freq is evenly spaced?
		for(valueColumnIndex=0; valueColumnIndex<numberOfValuesPerFreq; valueColumnIndex+=1)
			name = TouchstoneValueWaveName(fileInfo.numberOfPorts, fileInfo.parameterKind, fileInfo.format, valueColumnIndex)
			Wave w = $name
			SetScale /P x x0, dx, ""+fileInfo.frequencyUnit, w
		endfor
	endif
	
	for(valueColumnIndex=0; valueColumnIndex<numberOfValuesPerFreq; valueColumnIndex+=1)	// Set data units for value waves
		name = TouchstoneValueWaveName(fileInfo.numberOfPorts, fileInfo.parameterKind, fileInfo.format, valueColumnIndex)
		Wave w = $name
		String units
		strswitch(fileInfo.format)
			case "DB":
				if (valueColumnIndex & 1)
					units = "deg"
				else
					units = "DB"
				endif
				break
				
			case "MA":
				if (valueColumnIndex & 1)
					units = "deg"
				else
					units = ""									// I'm not sure what units should be here.
				endif
				break
				
			case "RI":
				if (valueColumnIndex & 1)
					units = ""									// I'm not sure what units should be here.
				else
					units = ""									// I'm not sure what units should be here.
				endif
				break
		endswitch
		SetScale /P d 0, 0, units, w
	endfor

	return 0
End

Function LoadTouchstoneFile(pathName, fileName, createTable, createGraph, graphMode, quiet)
	String pathName			// Igor symbolic path name or "" for dialog.
	String fileName				// file name, partial path and file name, or full path or "" for dialog.
	Variable createTable		// 0 = no table, 1 = edit even value column, 2 = edit odd value columns, 3 = edit both, 4 for dialog.
	Variable createGraph		// 0 = no graph, 1 = graph even value column, 2 = graph odd value columns, 3 = graph both, 4 for dialog.
	Variable graphMode			// kGraphModeAllInOne, kGraphModeValuesSeparate or kGraphModePairsSeparate, kGraphModeUndefined for dialog
	Variable quiet				// 0 = print diagnostic and error messages, 1 = print error messages only, 2 = do not print diagnostic or error messages
	
	Variable err
	
	if ( (createTable==4) || (createGraph==4) || (graphMode==kGraphModeUndefined))
		err = ViewTouchstoneFileDialog("Load Touchstone File", createTable, createGraph, graphMode)
		if (err != 0)
			return err
		endif
	endif

	// This puts up a dialog if the pathName and fileName do not specify a file.
	String message = "Select a Touchstone file"
	Variable refNum
	Open /R /Z=2 /P=$pathName /T="????" /M=message refNum as fileName	// /T flag presents "All files" in Open File dialog.
	
	// Save outputs from Open in a safe place.
	err = V_Flag
	String fullPath = S_fileName
	fileName = ParseFilePath(0, fullPath, ":", 1, 0)			// In case user chose a different file 	through dialog displayed by Open.

	if (err != 0)
		return err			// -1 means user canceled.
	endif

	STRUCT TouchstoneFileInfo fileInfo

	// Determine the number of ports from the file name extension
	String extension
	extension = ParseFilePath(4, fileName, ":", 0, 0)		// Should be s<n>p where <n> is the number of ports
	Variable len = strlen(extension)
	if (len < 3)
		err = -1
	else
		if (CmpStr(extension[0], "s") != 0)
			err = -1
		else
			if (CmpStr(extension[len-1], "p") != 0)
				err = -1
			else
				Variable numberOfPorts = str2num(extension[1,len-2])
				if (numberOfPorts<1 || numberOfPorts>100)
					err = -1
				else
					fileInfo.numberOfPorts = numberOfPorts
				endif
			endif
		endif
	endif
	if (err != 0)
		if (quiet != 2)
			Print "Bad file name extension - expected \".s<n>p\" where <n> is number of ports."
		endif
		Close refNum
		return -1
	endif

	// Load the option line information
	if (ExtractTouchstoneOptionInfo(refNum, fileInfo, quiet!=2) != 0)
		if (quiet != 2)
			Print "Error while loading option line of the file."
		endif
		Close refNum
		return -1
	endif
	
	Variable dataStartLine
	Variable dataStartPos			// Start of data as file position relative to start of file
	if (SkipOptionAndCommentLines(refNum, dataStartLine, dataStartPos, quiet!=2) != 0)
		if (quiet != 2)
			Print "Error while skipping option and comment lines in the file."
		endif
		Close refNum
		return -1
	endif
	
	// Create data folder for waves
	String dfName = ParseFilePath(3, fileName, ":", 0, 0)
	dfName = CleanupName(dfName, 0)
	NewDataFolder /O /S $dfName
	String fullDFPath = GetDataFolder(1)
	
	if (quiet == 0)
		Printf "Loading Touchstone data from \"%s\"\r", fullPath
	endif
	
	err = LoadTouchstoneData(fileInfo, refNum, dataStartPos, dataStartLine, quiet)

	Close refNum
	
	if (err == 0)
		StoreTouchstoneFileInfo(fileInfo)
	endif
	
	SetDataFolder ::					// Restore original current data folder
	
	if (err != 0)
		return err
	endif
	
	if (quiet == 0)
		Variable numWavesCreated = 2*fileInfo.numberOfPorts^2 + 1
		Printf "Created %d waves in data folder %s\r", numWavesCreated, GetDataFolder(1)+dfName
	endif

	if (createTable)
		Variable editEvenValueColumns = (createTable&1) != 0
		Variable editOddValueColumns = (createTable&2) != 0
		EditTouchstoneFileData(fullDFPath, editEvenValueColumns, editOddValueColumns)
	endif

	if (createGraph)
		Variable graphEvenValueColumns = (createGraph&1) != 0
		Variable graphOddValueColumns = (createGraph&2) != 0
		GraphTouchstoneFileData(fullDFPath, graphMode, graphEvenValueColumns, graphOddValueColumns)
	endif

	return 0			// Zero signifies no error.	
End

Function LoadTouchstoneFileMenuItem()
	LoadTouchstoneFile("", "", 4, 4, 4, 0)
End

static Function IsValidTouchstoneExtension(extension)
	String extension

	Variable isValidExtension = 0
	if (strlen(extension) == 3)							// TO DO: Change to allow extensions greater than three characters (e.g., s10p, s11p, etc.)
		if (CmpStr(extension[0], "s") == 0)
			if (CmpStr(extension[2], "p") == 0)
				isValidExtension = 1
			endif
		endif
	endif

	return isValidExtension
End

//	LoadAllTouchstoneFilesDialog(extension, createTable, createGraph)
//	Displays a dialog in which the user can enter parameters.
//	All of the parameters function both as inputs, to preset the dialog values,
//	and as output, to return the values to the calling routine.
//	The function result is 0 if OK or -1 if cancel.
Function LoadAllTouchstoneFilesDialog(extension, createTable, createGraph)
	String &extension					// Input and output
	Variable &createTable				// Input and output.
	Variable &createGraph				// Input and output.

	Variable isValidExtension = IsValidTouchstoneExtension(extension)

	if (!isValidExtension && CmpStr(extension,"????")!=0)
		extension = ".s2p"
	endif
	if ( (createTable<1) || (createTable>3) )
		createTable = 1
	endif
	if ( (createGraph<1) || (createGraph>3) )
		createGraph = 1
	endif
	
	// This is necessary because you can't pass an & parameter to Prompt.
	String extension2 = extension
	Variable createTable2 = createTable
	Variable createGraph2 = createGraph
	
	Prompt extension2, "Touchstone file name extension", popup ".s1p;.s2p;.s3p;.s4p;.s5p;.s6p;.s7p;.s8p;.s9p;Any extension"
	Prompt createTable2, "Create table", popup "No table;Even value columns;Odd value columns;All value columns;"
	Prompt createGraph2, "Create graph", popup "No graph;Even value columns;Odd value columns;All value columns;"
	DoPrompt "Load All Touchstone Files", extension2, createTable2, createGraph2
	if (V_Flag != 0)
		return -1			// User cancelled.
	endif
	
	if (CmpStr(extension2, "Any extension") == 0)
		extension2 = "????"
	endif
	
	extension = extension2
	createTable = createTable2 - 1
	createGraph = createGraph2 - 1
	
	return 0
End

Function GetTouchstoneColors(colorIndex, red, green, blue)
	Variable colorIndex
	Variable &red, &green, &blue						// Outputs
	
	colorIndex = mod(colorIndex, 8)					// Wrap around when we run out of colors.
	switch(colorIndex)
		case 0:
			red = 65535; green =  0; blue = 0					// Red
			break
			
		case 1:
			red = 2; green =  39321; blue = 1					// Green
			break

		case 2:
			red = 0; green =  0; blue = 65535					// Blue
			break
	
		case 3:
			red = 39321; green =  26208; blue = 1				// Brown
			break

		case 4:
			red = 65535; green =  32764; blue = 16385		// Orange
			break

		case 5:
			red = 0; green =  65535; blue = 0					// Lime
			break

		case 6:
			red = 52428; green =  1; blue = 41942				// Purple
			break

		case 7:
			red = 48059; green =  48059; blue = 48059		// Gray
			break
			
		case -1:												// This is used for frequency columns in tables.
		default:
			red = 0; green =  0; blue = 0						// Black
	endswitch
End

static Function EditTouchstoneData(tableName, tableTitle, w, numWavesInTable, colorIndex)
	String tableName				// Name to use for table
	String tableTitle
	Wave w							// Frequency or value wave
	Variable numWavesInTable
	Variable colorIndex
	
	if (numWavesInTable == 0)
		Edit /W=(5,42,678,460) w
		DoWindow/C $tableName
		DoWindow/T $tableName, tableTitle
		String userData = GetTouchstoneUserData()
		SetWindow kwTopWin, userdata(TouchstoneFileLoader)="VERSION:1.0;"	// Mark as Touchstone file loader window
	else
		AppendToTable w
	endif
	
	Variable red, green, blue
	GetTouchstoneColors(colorIndex, red, green, blue)
	
	ModifyTable rgb[numWavesInTable+1]=(red,green,blue)
End

Function EditTouchstoneFileData(fullDFPath, showEvenValueColumns, showOddValueColumns)
	String fullDFPath					// Full path to data folder containing Touchstone data.
	Variable showEvenValueColumns		// If true, shows DB, M or R data. The meaning of even and odd depends on the "format".
	Variable showOddValueColumns		// If true, shows angle or I data. The meaning of even and odd depends on the "format".
	
	if (!showEvenValueColumns && !showOddValueColumns)
		return 0
	endif
	
	if (!IsTouchstoneDataFolder(fullDFPath))
		return 0								// Data folder is not a Touchstone data folder.
	endif
	
	String saveDF = GetDataFolder(1)
	SetDataFolder fullDFPath
	
	STRUCT TouchstoneFileInfo fileInfo
	RetrieveTouchstoneFileInfo(fileInfo)		// Put info from global variables into structure
	
	String name
	Variable valueColumnIndex, numberOfValuesPerFreq
	Variable numWavesShown = 0
	
	String tableName, tableTitle
	tableName = CreateTouchstoneWindowName(2, 0, kGraphModeUndefined, fullDFPath, 0, tableTitle)
	
	name = TouchstoneFreqWaveName(fileInfo.frequencyUnit)
	Wave freq  = $name
	EditTouchstoneData(tableName, tableTitle, freq, numWavesShown, -1)			// -1 makes frequency column black
	numWavesShown += 1

	numberOfValuesPerFreq = (fileInfo.numberOfPorts ^ 2) * 2
	
	for(valueColumnIndex=0; valueColumnIndex<numberOfValuesPerFreq; valueColumnIndex+=1)
		Variable show = 0
		if (valueColumnIndex & 1)
			show =  showOddValueColumns
		else
			show = showEvenValueColumns
		endif
		
		if (show)
			name = TouchstoneValueWaveName(fileInfo.numberOfPorts, fileInfo.parameterKind, fileInfo.format, valueColumnIndex)
			Wave w = $name
			
			Variable colorIndex = trunc(valueColumnIndex/2)	// Even and odd value columns use same color.
			EditTouchstoneData(tableName, tableTitle, w, numWavesShown, colorIndex)

			numWavesShown += 1
		endif
	endfor
	
	SetDataFolder saveDF
	return numWavesShown
End

static Function/S CreateTouchstoneWindowName(windowType, isMultiFile, graphMode, fullDFPath, valueColumnIndex, title)
	Variable windowType				// 1 for graph, 2 for table
	Variable isMultiFile					// True if displaying data from multiple Touchstone files, false if displaying data from one Touchstone file.
	Variable graphMode					// kGraphModeAllInOne, kGraphModeValuesSeparate or kGraphModePairsSeparate. Applies only if isMultiFile is 0.
	String fullDFPath					// Full path to specific Touchstone data folder containing the data we are about to display
	Variable valueColumnIndex
	String &title						// OUTPUT: Window title
	
	String saveDF = GetDataFolder(1)
	SetDataFolder $fullDFPath
	NVAR numberOfPorts
	SVAR parameterKind
	SVAR format
	String dfName = GetDataFolder(0)
	SetDataFolder saveDF
	
	String name
	if (windowType == 1)
		name = "TG_"								// For "Touchstone Graph"
	else
		name = "TT_"								// For "Touchstone Table"
	endif

	String format2  = format
	if (!isMultiFile && graphMode==kGraphModePairsSeparate)
		format2 = ""								// Don't add (for example) "_DB" if we are displaying both DB and angle in the same graph.
	endif
	String valueName = TouchstoneValueWaveName(numberOfPorts, parameterKind, format2, valueColumnIndex)
	
	if (isMultiFile)
		name  += "MF_"							// For "multiple file"
		name += dfName
		name += "_" + valueName					// e.g., TG_MF_RX0934_S11_DB (assuming data folder name is RX0934)
		
		title = dfName + " - " + valueName			// e.g., RX0934-S11_DB (assuming data folder name is RX0934)
	else
		name  += "SF_"								// For "single file"
		name += dfName							// e.g., TG_SF_RX0934 (assuming data folder name is RX0934)
		title = dfName								// e.g., RX0934 (assuming data folder name is RX0934)

		switch(graphMode)
			case kGraphModeValuesSeparate:
			case kGraphModePairsSeparate:
				name += "_" + valueName			// e.g., TG_SF_RX0934_S11_DB (assuming data folder name is RX0934)
				title += " - " + valueName			// e.g., RX0934-S11_DB (assuming data folder name is RX0934)
				break
		endswitch
	endif
	
	if (strlen(name) > 31)							// Don't let name get too long
		name = name[0,30]
	endif
	
	// Make sure name is unique among windows. We assume that it does not conflict with other objects because we use a distinctive prefix.
	String list = WinList("*", ";", "")
	if (WhichListItem(name, list) >= 0)
		name = UniqueName(name, windowType+5, 0)	// This truncates name if it is too long.
	endif
	
	// Check title
	if (strlen(title) > 40)							// Don't let title get too long
		title = title[0,39]
	endif

	return name
End

Function TestTouchstoneWindowName(windowType, isMultiFile, graphMode, fullDFPath, valueColumnIndex)		// For testing only.
	Variable windowType				// 1 for graph, 2 for table
	Variable isMultiFile					// True if displaying data from multiple Touchstone files, false if displaying data from one Touchstone file.
	Variable graphMode
	String fullDFPath					// Full path to specific Touchstone data folder containing the data we are about to display
	Variable valueColumnIndex

	String name, title
	name = CreateTouchstoneWindowName(windowType, isMultiFile, graphMode, fullDFPath, valueColumnIndex, title)
	Printf "Window name: %s\r", name
	Printf "Window title: %s\r", title
End

static Function DisplayTouchstoneData(graphName, graphTitle, freq, w, createNewGraph, subWindowName, createNewSubwindow, valueColumnIndex, colorIndex, traceName)
	String graphName				// Name to use for graph
	String graphTitle
	Wave freq						// Frequency wave
	Wave w							// Value wave
	Variable createNewGraph
	String subWindowName			// Subwindow to display wave in or "" for main graph.
	Variable createNewSubwindow
	Variable valueColumnIndex
	Variable colorIndex
	String &traceName				// OUTPUT
	
	traceName = ""
	
	String userData = ""
	
	if (strlen(subWindowName) > 0)						// Using subwindows?
		if (createNewGraph)
			Display /W=(5,42,678,460)					// Create host window
			DoWindow/C $graphName
			DoWindow/T $graphName, graphTitle
			userData = GetTouchstoneUserData()
			SetWindow kwTopWin, userdata(TouchstoneFileLoader)="VERSION:1.0;"	// Mark as Touchstone file loader window
		endif
		if (createNewSubwindow)
			Variable left, top, right, bottom
			if (CmpStr(subWindowName,"Even")==0)
				left=0; top=0; right=.95; bottom=.5		// Even subwindow goes on top
			else
				left=0; top=.5; right=.95; bottom=1		// Odd subwindow goes on bottom
			endif
			if (leftx(w) == 0)
				Display /W=(left, top, right, bottom) /HOST=# w vs freq		// Frequencies not evenly spaced - must graph as XY pair.
			else
				Display /W=(left, top, right, bottom) /HOST=# w				// Frequencies are evenly spaced - can graph as waveform.
			endif
			RenameWindow #, $subWindowName
			// TO DO: Set window name based on data folder name
		else
			if (leftx(w) == 0)
				AppendToGraph w vs freq					// Frequencies not evenly spaced - must graph as XY pair.
			else
				AppendToGraph w							// Frequencies are evenly spaced - can graph as waveform.
			endif
		endif
	else
		// Not using subwindows
		if (createNewGraph)
			if (leftx(w) == 0)
				Display /W=(5,42,678,460) w vs freq		// Frequencies not evenly spaced - must graph as XY pair.
			else
				Display /W=(5,42,678,460) w			// Frequencies are evenly spaced - can graph as waveform.
			endif
			DoWindow/C $graphName
			DoWindow/T $graphName, graphTitle
			userData = GetTouchstoneUserData()
			SetWindow kwTopWin, userdata(TouchstoneFileLoader)="VERSION:1.0;"	// Mark as Touchstone file loader window
		else
			if (leftx(w) == 0)
				AppendToGraph w vs freq					// Frequencies not evenly spaced - must graph as XY pair.
			else
				AppendToGraph w							// Frequencies are evenly spaced - can graph as waveform.
			endif
		endif
	endif
	
	String list = TraceNameList("", ";", 1)
	Variable numTraces = ItemsInList(list)
	traceName = StringFromList(numTraces-1, list)		// Name of the trace we just added
	
	Variable lineStyle
	if (valueColumnIndex & 1)
	//	lineStyle = 3										// Dashed line for odd value columns. This is not needed now that we use a separate subwindow for angle data.
		lineStyle = 0										// Solid line for even value columns.
	else
		lineStyle = 0										// Solid line for even value columns.
	endif
	
	Variable red, green, blue
	GetTouchstoneColors(colorIndex, red, green, blue)
	
	ModifyGraph lstyle($traceName)=lineStyle, rgb($traceName)=(red,green,blue)
End

static Function ExecuteStyleMacro(macroName)
	String macroName										// Executes if macro exists
	
	if (Exists(macroName) == 5)							// Does such a macro exist?
		String cmd
		cmd = macroName + "()"
		Execute cmd
	endif
End

Function GraphTouchstoneFileData(fullDFPath, graphMode, showEvenValueColumns, showOddValueColumns)
	String fullDFPath					// Full path to data folder containing Touchstone data.
	Variable graphMode					// kGraphModeAllInOne, kGraphModeValuesSeparate or kGraphModePairsSeparate
	Variable showEvenValueColumns		// If true, shows DB, M or R data. The meaning of even and odd depends on the "format".
	Variable showOddValueColumns		// If true, shows angle or I data. The meaning of even and odd depends on the "format".
	
	if (!showEvenValueColumns && !showOddValueColumns)
		return 0
	endif
	
	if (!IsTouchstoneDataFolder(fullDFPath))
		return 0								// Data folder is not a Touchstone data folder.
	endif
	
	String saveDF = GetDataFolder(1)
	SetDataFolder fullDFPath
	String dfName = GetDataFolder(0)
	
	STRUCT TouchstoneFileInfo fileInfo
	RetrieveTouchstoneFileInfo(fileInfo)		// Put info from global variables into structure
	
	String name
	Variable valueColumnIndex, numberOfValuesPerFreq
	Variable numWavesShown, numEvenWavesShown, numOddWavesShown
	String subWindowName = ""					// When displaying even and odd columns (e.g., db and angle) we use two subwindows.
	String legendText
	
	name = TouchstoneFreqWaveName(fileInfo.frequencyUnit)
	Wave freq  = $name

	numberOfValuesPerFreq = (fileInfo.numberOfPorts ^ 2) * 2
	
	numWavesShown = 0
	numEvenWavesShown = 0
	numOddWavesShown = 0
	
	Variable colorIndex
	String traceName			// Set by DisplayTouchstoneData.
	String graphName, graphTitle
	String styleMacroName =  ""
	String temp
	
	Variable createNewGraph = 0, createNewSubwindow = 0
	String graphList = ""
	
	if (graphMode == kGraphModePairsSeparate)						// This eliminates an inconsistent combination of parameters
		if (showEvenValueColumns==0 || showOddValueColumns==0)	// Not showing pairs?
			graphMode = kGraphModeValuesSeparate
		endif
	endif
	
	if (showEvenValueColumns)
		subWindowName = ""
		if (showOddValueColumns && graphMode!=kGraphModeValuesSeparate)
			subWindowName = "Even"
		endif
		legendText = ""
		for(valueColumnIndex=0; valueColumnIndex<numberOfValuesPerFreq; valueColumnIndex+=2)
			name = TouchstoneValueWaveName(fileInfo.numberOfPorts, fileInfo.parameterKind, fileInfo.format, valueColumnIndex)
			Wave w = $name
			colorIndex = trunc(valueColumnIndex/2)	// Even and odd value columns use different dash pattern, same color.
			
			switch(graphMode)
				case kGraphModeAllInOne:
					createNewGraph = numWavesShown == 0			
					createNewSubwindow = numEvenWavesShown == 0
					styleMacroName = "TouchstoneGraphStyleSingle1"	
					break
				
				case kGraphModeValuesSeparate:
					createNewGraph = 1			
					createNewSubwindow = 1			
					styleMacroName = "TouchstoneGraphStyleSingle2"
					break
					
				case kGraphModePairsSeparate:
					createNewGraph = 1			
					createNewSubwindow = 1			
					styleMacroName = "TouchstoneGraphStyleSingle3"
					break
			endswitch
			
			if (createNewGraph)
				graphName = CreateTouchstoneWindowName(1, 0, graphMode, fullDFPath, valueColumnIndex, graphTitle)
				graphList  += graphName + ";"
			endif
			
			DisplayTouchstoneData(graphName, graphTitle, freq, w, createNewGraph, subWindowName, createNewSubwindow, valueColumnIndex, colorIndex, traceName)
			numWavesShown += 1
			numEvenWavesShown += 1
			
			sprintf temp, "\\s(%s) %s:%s\r", traceName, dfName, NameOfWave(w)
			legendText += temp
			if (graphMode!=kGraphModeAllInOne || numEvenWavesShown==numberOfValuesPerFreq/2)
				legendText = RemoveEnding(legendText)				// Remove last CR.
				Legend/N=legend0/A=RT/X=5/Y=5 legendText
				legendText = ""
			endif
			
			if (graphMode != kGraphModeAllInOne)
				ExecuteStyleMacro(styleMacroName)
			endif
		endfor
		
		if (graphMode == kGraphModeAllInOne)
			ExecuteStyleMacro(styleMacroName)
		endif
	endif
	
	if (showOddValueColumns)
		subWindowName = ""
		if (showEvenValueColumns && graphMode!=kGraphModeValuesSeparate)
			subWindowName = "Odd"
			SetActiveSubwindow ##					// Reactivate host instead of the Even subwindow.
		endif
		legendText = ""
		for(valueColumnIndex=1; valueColumnIndex<numberOfValuesPerFreq; valueColumnIndex+=2)
			name = TouchstoneValueWaveName(fileInfo.numberOfPorts, fileInfo.parameterKind, fileInfo.format, valueColumnIndex)
			Wave w = $name
			colorIndex = trunc(valueColumnIndex/2)	// Even and odd value columns use different dash pattern, same color.

			switch(graphMode)
				case kGraphModeAllInOne:
					createNewGraph = numWavesShown == 0			
					createNewSubwindow = numOddWavesShown == 0			
					styleMacroName = "TouchstoneGraphStyleSingle1"	
					break
				
				case kGraphModeValuesSeparate:
					createNewGraph = 1			
					createNewSubwindow = 1			
					styleMacroName = "TouchstoneGraphStyleSingle2"	
					break
				
				case kGraphModePairsSeparate:
					createNewGraph = 0			
					createNewSubwindow = 1
					String targetGraph = StringFromList(numOddWavesShown, graphList)	// Activate the graph in which this odd value should be shown.
					DoWindow/F $targetGraph
					SetActiveSubwindow $targetGraph										// Activate main graph so this subwindow goes into it.
					styleMacroName = "TouchstoneGraphStyleSingle3"	
					break
			endswitch

			if (createNewGraph)
				graphName = CreateTouchstoneWindowName(1, 0, graphMode, fullDFPath, valueColumnIndex, graphTitle)
				graphList  += graphName + ";"
			endif
			
			DisplayTouchstoneData(graphName, graphTitle, freq, w, createNewGraph, subWindowName, createNewSubwindow, valueColumnIndex, colorIndex, traceName)
			numWavesShown += 1
			numOddWavesShown += 1
			
			sprintf temp, "\\s(%s) %s:%s\r", traceName, dfName, NameOfWave(w)
			legendText += temp
			if (graphMode!=kGraphModeAllInOne || numOddWavesShown==numberOfValuesPerFreq/2)
				legendText = RemoveEnding(legendText)				// Remove last CR.
				Legend/N=legend0/A=RT/X=5/Y=5 legendText
				legendText = ""
			endif
			
			if (graphMode != kGraphModeAllInOne)
				ExecuteStyleMacro(styleMacroName)
			endif

			if (graphMode != kGraphModeAllInOne)
				SetActiveSubwindow $(WinName(0, 1))	// Activate the main graph. This is just to get rid of the distracting yellow-dashed line indicating the active subwindow.
			endif
		endfor

		if (graphMode == kGraphModeAllInOne)
			ExecuteStyleMacro(styleMacroName)
		endif
	endif
	
	if (graphMode == kGraphModeAllInOne)
		SetActiveSubwindow $(WinName(0, 1))			// Activate the main graph. This is just to get rid of the distracting yellow-dashed line indicating the active subwindow.
	endif
	
	if (graphMode != kGraphModeAllInOne)
		// Fix order of graphs so they will be in intuitive order.
		Variable index
		for(index = ItemsInList(graphList)-1; index>= 0; index-=1)
			graphName = StringFromList(index, graphList)
			DoWindow/F $graphName
		endfor
	endif
	
	SetDataFolder saveDF
	return numWavesShown
End

Function ViewTouchstoneFileMenuItem()
	Variable createTable, createGraph, graphMode
	if (ViewTouchstoneFileDialog("View Touchstone Data From One File", createTable, createGraph, graphMode) != 0)
		return -1
	endif
	
	if (createTable==0 && createGraph==0)
		return 0
	endif
	
	String fullDFPath = GetDataFolder(1)

	Variable numWavesShown =  0
	
	if (createTable)
		Variable editEvenValueColumns = (createTable&1) != 0
		Variable editOddValueColumns = (createTable&2) != 0
		numWavesShown += EditTouchstoneFileData(fullDFPath, editEvenValueColumns, editOddValueColumns)
	endif

	if (createGraph)
		Variable graphEvenValueColumns = (createGraph&1) != 0
		Variable graphOddValueColumns = (createGraph&2) != 0
		numWavesShown += GraphTouchstoneFileData(fullDFPath, graphMode, graphEvenValueColumns, graphOddValueColumns)
	endif
	
	if (numWavesShown == 0)
		DoAlert 0, "No Touchstone data found. The current data folder must contain Touchstone waves and variables."
	endif
End

Function LoadAllTouchstoneFiles(pathName, extension, createTable, createGraph, quiet)
	String pathName			// Name of an Igor symbolic folder created by Misc->NewPath
	String extension			// e.g., ".sp2", ".sp3" or "????" for all files
	Variable createTable		// 1 = Yes, 2 = No, 0 for dialog.
	Variable createGraph		// 0 = no graph, 1 = graph even value column, 2 = graph odd value columns, 3 = graph both, 4 for dialog.
	Variable quiet				// 0 = print diagnostic and error messages, 1 = print error messages only, 2 = do not print diagnostic or error messages

	if (!IsValidTouchstoneExtension(extension) || createTable==4 || createGraph==4)
		if (LoadAllTouchstoneFilesDialog(extension, createTable, createGraph) != 0)
			return -1					// User cancelled.
		endif
	endif
	
	if (strlen(pathName) == 0)
		String message
		message = "Select a directory containing Touchstone files"
		NewPath/O/M=message/Q TouchstoneDataPath		// This displays a dialog in which you can select a folder
		if (V_flag != 0)
			return V_flag									// -1 means user canceled
		endif
		pathName = "TouchstoneDataPath"
	endif
	
	PathInfo $pathName
	String fullPath = S_path								// Full path to directory. This will end with trailing colon.
	fullPath = RemoveEnding(fullPath, ":")					// Remove trailing colon - otherwise ParseFilePath does not work right, at least in Igor Pro 5.04B.
	String dfName = ParseFilePath(3, fullPath, ":", 1, 0)
	dfName = CleanupName(dfName, 0)
	
	String saveDF =  GetDataFolder(1)
	NewDataFolder /O /S $dfName
	String fullDFPath = GetDataFolder(1)
	
	if (quiet == 0)
		Printf "Loading Touchstone data from directory \"%s\"\r", fullPath
	endif

	Variable err = 0

	String fileName
	Variable fileIndex = 0
	do
		fileName = IndexedFile($pathName, fileIndex, extension)
		if (strlen(fileName) == 0)
			break											// No more files
		endif
	
		err = LoadTouchstoneFile(pathName, fileName, 0, 0, 1, quiet)
		if (err != 0)
			return err
		endif
	
		fileIndex += 1
	while(1)

	if (createTable)
		Variable editEvenValueColumns = (createTable&1) != 0
		Variable editOddValueColumns = (createTable&2) != 0
		EditTouchstoneFilesData(fullDFPath, editEvenValueColumns, editOddValueColumns)
	endif
	
	if (createGraph)
		Variable graphEvenValueColumns = (createGraph&1) != 0
		Variable graphOddValueColumns = (createGraph&2) != 0
		GraphTouchstoneFilesData(fullDFPath, graphEvenValueColumns, graphOddValueColumns)
	endif
	
	SetDataFolder saveDF
	
	return 0
End

Function LoadAllTouchstoneFilesMenuItem()
	LoadAllTouchstoneFiles("", "", 4, 4, 0)
End

static Function IsTouchstoneDataFolder(fullDFPath)	// Returns true of specified data folder is a Touchstone data folder.
	String fullDFPath
	
	String saveDF = GetDataFolder(1)
	SetDataFolder fullDFPath
	
	Variable result = 0
	
	SVAR/Z frequencyUnit = frequencyUnit
	NVAR/Z numberOfPorts = numberOfPorts
	SVAR/Z parameterKind = parameterKind
	SVAR/Z format = format
	
	if (SVAR_Exists(frequencyUnit))
		if (NVAR_Exists(numberOfPorts))
			if (SVAR_Exists(parameterKind))
				if (SVAR_Exists(format))
					result = 1
				endif
			endif
		endif
	endif
	
	SetDataFolder(saveDF)
	return result
End

static Function/S GetTouchstoneSubDataFolderList(fullMasterDFPath, numberOfPorts)	// Returns list of Touchstone data folders within the current master data folder.
	String fullMasterDFPath			// Full path to master data folder containing Touchstone sub-data folders
	Variable &numberOfPorts			 // Output.

	String list = ""
	numberOfPorts = 0
	
	String saveDF = GetDataFolder(1)
	SetDataFolder fullMasterDFPath
	
	Variable index = 0
	do
		String subDFName = GetIndexedObjName(fullMasterDFPath, 4, index)
		if (strlen(subDFName) == 0)
			break
		endif
		
		if (IsTouchstoneDataFolder(fullMasterDFPath+subDFName))
			NVAR thisNumberOfPorts = :$(subDFName):numberOfPorts
			if (numberOfPorts == 0)			// Have not yet found the first sub-data folder?
				numberOfPorts = thisNumberOfPorts
				list += subDFName + ";"
			else
				if (thisNumberOfPorts == numberOfPorts)		// Matches number of ports we are looking for?
					list += subDFName + ";"
				endif
			endif		
		endif
		
		index += 1
	while(1)

	SetDataFolder saveDF
	
	return list
End

Function EditTouchstoneFilesData(fullMasterDFPath, showEvenValueColumns, showOddValueColumns)	// Returns number of Touchstone data folders processed
	String fullMasterDFPath			// Full path to master data folder containing Touchstone data folders for individual files.
	Variable showEvenValueColumns		// If true, shows DB, M or R data. The meaning of even and odd depends on the "format".
	Variable showOddValueColumns		// If true, shows angle or I data. The meaning of even and odd depends on the "format".
	
	if (!showEvenValueColumns && !showOddValueColumns)
		return 0
	endif
	
	Variable masterNumberOfPorts
	String subDataFolderList = GetTouchstoneSubDataFolderList(fullMasterDFPath, masterNumberOfPorts)
	if (masterNumberOfPorts<1 || strlen(subDataFolderList)==0)
		return 0						// No Touchstone data folders processed
	endif
	
	String saveDF = GetDataFolder(1)
	SetDataFolder fullMasterDFPath
	
	String subDFName
	
	Variable numberOfValuesPerFreq = (masterNumberOfPorts ^ 2) * 2
	Variable valueColumnIndex
	String name
	Variable numTablesMade = 0
	String tableName, tableTitle
	
	// Loop through wave for each value column. Loop backwards so tables will be in intuitive order.
	for(valueColumnIndex=numberOfValuesPerFreq-1; valueColumnIndex>=0; valueColumnIndex-=1)
		Variable numWavesShown = 0

		Variable show = 0
		if (valueColumnIndex & 1)
			show =  showOddValueColumns
		else
			show = showEvenValueColumns
		endif
		
		if (show)
			Variable numSubDataFolders = ItemsInList(subDataFolderList)
			Variable subDFIndex

			// Pick one wave from each sub-data folder
			for (subDFIndex=0; subDFIndex<numSubDataFolders; subDFIndex+=1)		// Loop through sub-data folders.
				subDFName = StringFromList(subDFIndex, subDataFolderList)

				SVAR frequencyUnit = :$(subDFName):frequencyUnit
				NVAR numberOfPorts = :$(subDFName):numberOfPorts
				SVAR parameterKind = :$(subDFName):parameterKind
				SVAR format = :$(subDFName):format

				Variable colorIndex = subDFIndex							// Waves from same data folder use same color.

				name = TouchstoneFreqWaveName(frequencyUnit)
				Wave freq  = :$(subDFName):$name

				if (numWavesShown == 0)
					tableName = CreateTouchstoneWindowName(2, 1, kGraphModeUndefined, fullMasterDFPath+subDFName, valueColumnIndex, tableTitle)
				endif
				EditTouchstoneData(tableName, tableTitle, freq, numWavesShown, colorIndex)
				numWavesShown += 1

				name = TouchstoneValueWaveName(numberOfPorts, parameterKind, format, valueColumnIndex)
				Wave w = :$(subDFName):$name
				EditTouchstoneData(tableName, tableTitle, w, numWavesShown, colorIndex)
				numWavesShown += 1
			endfor
			numTablesMade += 1
		endif
	endfor
	
	SetDataFolder saveDF
	return numTablesMade
End

Function GraphTouchstoneFilesData(fullMasterDFPath, showEvenValueColumns, showOddValueColumns)	// Returns number of Touchstone data folders processed
	String fullMasterDFPath			// Full path to master data folder containing Touchstone data folders for individual files.
	Variable showEvenValueColumns		// If true, shows DB, M or R data. The meaning of even and odd depends on the "format".
	Variable showOddValueColumns		// If true, shows angle or I data. The meaning of even and odd depends on the "format".
	
	if (!showEvenValueColumns && !showOddValueColumns)
		return 0
	endif
	
	Variable masterNumberOfPorts
	String subDataFolderList = GetTouchstoneSubDataFolderList(fullMasterDFPath, masterNumberOfPorts)
	if (masterNumberOfPorts<1 || strlen(subDataFolderList)==0)
		return 0						// No Touchstone data folders processed
	endif
	
	String saveDF = GetDataFolder(1)
	SetDataFolder fullMasterDFPath
	
	String subDFName
	
	Variable numberOfValuesPerFreq = (masterNumberOfPorts ^ 2) * 2
	Variable valueColumnIndex
	String name
	Variable numGraphsMade = 0
	String graphName, graphTitle
	
	// Loop through wave for each value column. Loop backwards so graphs will be in intuitive order
	for(valueColumnIndex=numberOfValuesPerFreq-1; valueColumnIndex>=0; valueColumnIndex-=1)
		Variable numWavesShown = 0
		String legendText = ""

		Variable show = 0
		if (valueColumnIndex & 1)
			show =  showOddValueColumns
		else
			show = showEvenValueColumns
		endif
		
		if (show)
			Variable numSubDataFolders = ItemsInList(subDataFolderList)
			Variable subDFIndex

			// Pick one wave from each sub-data folder
			for (subDFIndex=0; subDFIndex<numSubDataFolders; subDFIndex+=1)		// Loop through sub-data folders.
				subDFName = StringFromList(subDFIndex, subDataFolderList)

				SVAR frequencyUnit = :$(subDFName):frequencyUnit
				NVAR numberOfPorts = :$(subDFName):numberOfPorts
				SVAR parameterKind = :$(subDFName):parameterKind
				SVAR format = :$(subDFName):format

				name = TouchstoneFreqWaveName(frequencyUnit)
				Wave freq  = :$(subDFName):$name

				name = TouchstoneValueWaveName(numberOfPorts, parameterKind, format, valueColumnIndex)
				Wave w = :$(subDFName):$name
				Variable colorIndex = subDFIndex	// Even and odd value columns use different dash pattern. Waves from same data folder use same color.
				String traceName
				if (numWavesShown == 0)
					graphName = CreateTouchstoneWindowName(1, 1, kGraphModeUndefined, fullMasterDFPath+subDFName, valueColumnIndex, graphTitle)
				endif
				DisplayTouchstoneData(graphName, graphTitle, freq, w, numWavesShown==0, "",  0, valueColumnIndex==0, colorIndex, traceName)
				String temp
				sprintf temp, "\\s(%s) %s:%s\r", traceName, subDFName, NameOfWave(w)
				legendText += temp
				numWavesShown += 1
			endfor
	
			if (numWavesShown > 0)
				legendText = RemoveEnding(legendText)				// Remove last CR.
				Legend/N=legend0/A=RT/X=5/Y=5 legendText
				numGraphsMade += 1
				ExecuteStyleMacro("TouchstoneGraphStyleMulti1")
			endif
		endif
	endfor
	
	SetDataFolder saveDF
	
	return numGraphsMade
End

//	ViewTouchstoneFilesDialog(title, createTable, createGraph)
//	Displays a dialog in which the user can enter parameters.
//	This is called when we are displaying Touchstone data from multiple files.
//	All of the parameters function both as inputs, to preset the dialog values,
//	and as output, to return the values to the calling routine.
//	The function result is 0 if OK or -1 if cancel.
Function ViewTouchstoneFilesDialog(title, createTable, createGraph)
	String title							// Dialog title
	Variable &createTable				// Input and output. 0 = no table, 1 = edit even value column, 2 = edit odd value columns, 3 = edit both
	Variable &createGraph				// Input and output. 0 = no graph, 1 = graph even value column, 2 = graph odd value columns, 3 = graph both

	Variable graphMode
	GetTouchstoneViewPrefs(createTable, createGraph, graphMode)
	
	Variable createTable2 = createTable + 1
	Variable createGraph2 = createGraph + 1
	
	Prompt createTable2, "Create table", popup "No table;Even value columns;Odd value columns;All value columns;"
	Prompt createGraph2, "Create graph", popup "No graph;Even value columns;Odd value columns;All value columns;"
	DoPrompt title, createTable2, createGraph2
	if (V_Flag != 0)
		return -1			// User cancelled.
	endif
	
	createTable = createTable2 - 1
	createGraph = createGraph2 - 1

	SetTouchstoneViewPrefs(createTable, createGraph, graphMode)
	
	return 0
End

Function ViewTouchstoneFilesMenuItem()
	Variable createTable, createGraph
	if (ViewTouchstoneFilesDialog("View Touchstone Data From Multiple Files", createTable, createGraph) != 0)
		return -1
	endif
	
	if (createTable==0 && createGraph==0)
		return 0
	endif
	
	String fullDFPath = GetDataFolder(1)

	Variable numWindowsShown =  0
	
	if (createTable)
		Variable editEvenValueColumns = (createTable&1) != 0
		Variable editOddValueColumns = (createTable&2) != 0
		numWindowsShown += EditTouchstoneFilesData(fullDFPath, editEvenValueColumns, editOddValueColumns)
	endif

	if (createGraph)
		Variable graphEvenValueColumns = (createGraph&1) != 0
		Variable graphOddValueColumns = (createGraph&2) != 0
		numWindowsShown += GraphTouchstoneFilesData(fullDFPath, graphEvenValueColumns, graphOddValueColumns)
	endif
	
	if (numWindowsShown == 0)
		DoAlert 0, "No Touchstone data found. Set current data folder to master Touchstone data folder containing Touchstone data folders for individual files."
	endif
End

static Function/S GetTouchstoneUserData()
	String userData = ""
	String temp
	sprintf temp, "VERSION:%.2f;", kTouchStoneFileLoaderVersion
	userData += temp
	return userData
End

Function KillAllTouchstoneWindows()
	String list = WinList("*", ";", "WIN:3")					// Graphs and tables
	Variable numWindows = ItemsInList(list)
	Variable index
	
	for(index=numWindows-1; index>=0; index-=1)
		String name = StringFromList(index, list)
		String userData = GetUserData(name, "", "TouchstoneFileLoader")
		if (strlen(userData) != 0)
			DoWindow/K $name
		endif	
	endfor
End
